#include <cxxtest/TestSuite.h>

#include "HeapData.h"
#include "MockErrno.h"
#include "MockSelector.h"
#include "MockSocket.h"
#include "RealMalloc.h"
#include "SelectableSocket.h"

class MockSocketDataHandlerForTests : public ISelectableSocket::ISocketDataHandler
{
public:
    MockSocketDataHandlerForTests()
    {
        handleDataCalled = false;
    }

    HeapData lastData;
    bool handleDataCalled;

    void handleData( const ConstSimpleData &data )
    {
        handleDataCalled = true;
        lastData = data;
    }
};

class MockForSelectableSocket :
	public RealMalloc,
    public MockErrno
{
public:
};

class SelectableSocketTestSuite : public CxxTest::TestSuite
{
public:
	MockForSelectableSocket *mock;
    MockSocketDataHandlerForTests *dataHandler;
    MockSocket *socket;
    MockSelector *selector;

	void setUp()
	{
		TS_ASSERT( mock = new MockForSelectableSocket );
        TS_ASSERT( dataHandler = new MockSocketDataHandlerForTests );
        TS_ASSERT( socket = new MockSocket );
        TS_ASSERT( selector = new MockSelector );
	}

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete selector );
		TS_ASSERT_THROWS_NOTHING( delete socket );
		TS_ASSERT_THROWS_NOTHING( delete dataHandler );
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_constructor_destructor()
    {
        SelectableSocket *selectableSocket;

        TS_ASSERT( selectableSocket = new SelectableSocket( *socket, *selector, *dataHandler ) );

        TS_ASSERT( selector->registerWasCalled );
        TS_ASSERT_EQUALS( selector->lastRegisteredSelectable, selectableSocket );
        TS_ASSERT_EQUALS( selector->lastRegistetedDescriptor, socket->lastDescriptor );

        TS_ASSERT_THROWS_NOTHING( delete selectableSocket );

        TS_ASSERT( selector->unregisterWasCalled );
        TS_ASSERT_EQUALS( selector->lastUnregisteredSelectable, selectableSocket );
        TS_ASSERT_EQUALS( selector->lastUnregistetedDescriptor, socket->lastDescriptor );
    }

    void test_can_read()
    {
        SelectableSocket *selectableSocket;

        TS_ASSERT( selectableSocket = new SelectableSocket( *socket, *selector, *dataHandler ) );

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };

        socket->nextReceivedData.append( ConstSimpleData( buffer, sizeof(buffer) ) );

        TS_ASSERT_THROWS_NOTHING( selectableSocket->canRead() );

        TS_ASSERT_EQUALS( dataHandler->lastData.size(), sizeof(buffer) );
        TS_ASSERT_SAME_DATA( dataHandler->lastData.data(), buffer, sizeof(buffer) );

        TS_ASSERT_THROWS_NOTHING( delete selectableSocket );
    }

    void test_cast_to_socket()
    {
        SelectableSocket *selectableSocket;

        TS_ASSERT( selectableSocket = new SelectableSocket( *socket, *selector, *dataHandler ) );

        ISocket &asSocket( *selectableSocket );

        TS_ASSERT_EQUALS( &asSocket, socket );

        TS_ASSERT_THROWS_NOTHING( delete selectableSocket );
    }

    void test_can_read__socket_closes()
    {
        SelectableSocket *selectableSocket;

        TS_ASSERT( selectableSocket = new SelectableSocket( *socket, *selector, *dataHandler ) );

        socket->socketShouldClose = true;

        TS_ASSERT_THROWS_NOTHING( selectableSocket->canRead() );

        TS_ASSERT( selector->unregisterWasCalled );
        TS_ASSERT_EQUALS( selector->lastUnregisteredSelectable, selectableSocket );
        TS_ASSERT_EQUALS( selector->lastUnregistetedDescriptor, socket->lastDescriptor );

        TS_ASSERT( !dataHandler->handleDataCalled );

        socket->receiveWasCalled = false;
        TS_ASSERT_THROWS_NOTHING( selectableSocket->canRead() );
        TS_ASSERT( !socket->receiveWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete selectableSocket );
    }

    void test_close_socket()
    {
        SelectableSocket *selectableSocket;

        TS_ASSERT( selectableSocket = new SelectableSocket( *socket, *selector, *dataHandler ) );

        TS_ASSERT_THROWS_NOTHING( selectableSocket->closeSocket() );

        TS_ASSERT( selector->unregisterWasCalled );
        TS_ASSERT_EQUALS( selector->lastUnregisteredSelectable, selectableSocket );
        TS_ASSERT_EQUALS( selector->lastUnregistetedDescriptor, socket->lastDescriptor );

        TS_ASSERT( socket->closeSocketWasCalled );

        TS_ASSERT_THROWS_NOTHING( selectableSocket->canRead() );
        TS_ASSERT( !socket->receiveWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete selectableSocket );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
