#ifndef __Weaver_h__
#define __Weaver_h__

#include "AutoThreadListener.h"
#include "Event.h"
#include "EventSets.h"
#include "IWeaver.h"
#include "Map.h"
#include "Oracle.h"
#include "Vector.h"

class Weaver : public IWeaver
{
public:
    Weaver( ISelector &selector ) : _selector( selector )
    {
    }

    ~Weaver()
    {
        discardListeners();
    }

	void handleAcceptedSocket( ISocket &socket )
	{
        _threadListeners.append( T::createThreadListener( *this, _selector, socket ) );
    }

    void removeThread( IThreadListener *thread )
    {
        removeThreadFromDataStructures( thread );
        destroyThread( thread );

        if ( everyoneSynchornized() )
        {
            selectNextEvent();
        }
    }

    void bSync( IThreadListener *thread,
                const Vector<Event> &requested,
                const Vector<Event> &watched,
                const Vector<Event> &blocked )
    {
        storeEvents( thread, requested, watched, blocked );

        if ( everyoneSynchornized() )
        {
            selectNextEvent();
        }
    }

private:
    ISelector &_selector;
    Vector<IThreadListener *> _threadListeners;
    Map<IThreadListener *, EventSets> _threadsToEvents;
    Vector<Event *> _requested;
    Vector<Event *> _blocked;
    Vector<Event> _enabled;
    Event _selectedEvent;

    bool everyoneSynchornized() const
    {
        return _threadsToEvents.size() == _threadListeners.size();
    }

    void storeEvents( IThreadListener *thread,
                      const Vector<Event> &requested,
                      const Vector<Event> &watched,
                      const Vector<Event> &blocked )
    {
        _threadsToEvents[thread] = EventSets( requested, watched, blocked );
    }

    void discardListeners()
    {
        for ( unsigned i = 0; i < _threadListeners.size(); ++i )
        {
            T::discardThreadListener( _threadListeners[i] );
        }
    }

    void collectRequested()
    {
        Map<IThreadListener *, EventSets>::iterator it;

        for ( it = _threadsToEvents.begin(); it != _threadsToEvents.end(); ++it )
        {
            for ( unsigned i = 0; i < it->second._requested.size(); ++i )
            {
                _requested.append( &(it->second._requested[i]) );
            }
        }
    }

    void collectBlocked()
    {
        Map<IThreadListener *, EventSets>::iterator it;

        for ( it = _threadsToEvents.begin(); it != _threadsToEvents.end(); ++it )
        {
            for ( unsigned i = 0; i < it->second._blocked.size(); ++i )
            {
                _blocked.append( &(it->second._blocked[i]) );
            }
        }
    }

    void selectEvent()
    {
        collectRequested();
        collectBlocked();
        collectEnabled();

        chooseEnabled();
    }

    void collectEnabled()
    {
        for ( unsigned i = 0; i < _requested.size(); ++i )
        {
            if ( enabledEvent( _requested[i] ) )
            {
                _enabled.append( *_requested[i] );
            }
        }
    }

    void chooseEnabled()
    {
        if ( _enabled.empty() )
        {
            throw Error( Error::SELECTION_DEADLOCKED );
        }

        _selectedEvent = Oracle::choose( _enabled );
    }

    bool enabledEvent( Event *event )
    {
        for ( unsigned i = 0; i < _blocked.size(); ++i )
        {
            if ( *_blocked[i] == *event )
            {
                return false;
            }
        }

        return true;
    }

    bool threadRequestedOrWatched( IThreadListener *thread )
    {
        return threadRequested( thread ) || threadWatched( thread );
    }

    bool threadRequested( IThreadListener *thread )
    {
        for ( unsigned i = 0; i < _threadsToEvents[thread]._requested.size(); ++i )
        {
            if ( _threadsToEvents[thread]._requested[i].type() == _selectedEvent.type() )
            {
                return true;
            }
        }

        return false;
    }

    bool threadWatched( IThreadListener *thread )
    {
        for ( unsigned i = 0; i < _threadsToEvents[thread]._watched.size(); ++i )
        {
            if ( _threadsToEvents[thread]._watched[i].type() == _selectedEvent.type() )
            {
                return true;
            }
        }

        return false;
    }

    void announceEvent()
    {
        for ( unsigned i = 0; i < _threadListeners.size(); ++i )
        {
            if ( threadRequestedOrWatched( _threadListeners[i] ) )
            {
                _threadListeners[i]->announceChosenEvent( _selectedEvent );
                _threadsToEvents.erase( _threadListeners[i] );
            }
        }
    }

    void clearDataStructures()
    {
        _requested.clear();
        _blocked.clear();
        _enabled.clear();
    }

    void removeThreadFromDataStructures( IThreadListener *thread )
    {
        if ( _threadListeners.exists( thread ) )
        {
            _threadListeners.erase( thread );
        }
        if ( _threadsToEvents.exists( thread ) )
        {
            _threadsToEvents.erase( thread );
        }
    }

    void destroyThread( IThreadListener *thread )
    {
        T::discardThreadListener( thread );
    }

    void selectNextEvent()
    {
        selectEvent();
        announceEvent();
        clearDataStructures();
    }
};

#endif // __Weaver_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
