#ifndef __AbstractSafetyPatcher_h__
#define __AbstractSafetyPatcher_h__

#include "AbstractionData.h"
#include "DirectAbstraction.h"
#include "EventVector.h"
#include "NFA.h"
#include "SafetyPatch.h"

class AbstractSafetyPatcher
{
public:
    static SafetyPatch patch( NFA &automaton, AbstractionData &moduleBasedAbstraction )
    {
        return AbstractSafetyPatcher( automaton, moduleBasedAbstraction ).run();
    }

    static SafetyPatch patch( NFA &automaton, AbstractionData &moduleBasedAbstraction, DirectAbstraction &directAbstraction )
    {
        return AbstractSafetyPatcher( automaton, moduleBasedAbstraction, directAbstraction ).run();
    }

private:
    typedef Vector<String> StateSet;

    NFA &_automaton;
    DirectAbstraction *_directAbstraction;
    unsigned _predecessorIndex;
    Vector<Event> _allEvents;

    SafetyPatch _patch;

    UniqueVector<String> _bad;
    UniqueVector<String> _badFrontier;
    UniqueVector<String> _predecessorsOfBad;

    String _newBadState;
    StateSet _newPredecessors;

    bool _deleteDirectAbstraction;

    AbstractSafetyPatcher( NFA &automaton, AbstractionData &moduleBasedAbstraction, DirectAbstraction &directAbstraction ) :
        _automaton( automaton ), _directAbstraction( &directAbstraction )
    {
        _deleteDirectAbstraction = false;

        _allEvents = moduleBasedAbstraction._composedGraph->_events;

        TS_TRACE( "Starting to patch -- Dumping automaton" );
        _automaton.dump();
    }

    AbstractSafetyPatcher( NFA &automaton, AbstractionData &moduleBasedAbstraction ) :
        _automaton( automaton )
    {
        _directAbstraction = new DirectAbstraction( automaton, moduleBasedAbstraction );
        _deleteDirectAbstraction = true;

        _allEvents = moduleBasedAbstraction._composedGraph->_events;

        TS_TRACE( "Abstract Patcher starting" );

        // TS_TRACE( "Dumping automaton" );
        // _automaton.dump();
    }

    ~AbstractSafetyPatcher()
    {
        if ( _deleteDirectAbstraction )
        {
            delete _directAbstraction;
            _directAbstraction = NULL;
            _deleteDirectAbstraction = false;
        }
    }

    SafetyPatch run()
    {
        initializeBad();
        initializePredecessors();

        //        dumpFrontier();

        while ( needToRefineAPredecessor() )
        {
            refineAPredecessor();
            //            dumpFrontier();
        }

        TS_TRACE( "No further refinement needed!\n" );

        // TS_TRACE( "*** Automaton After: ***" );
        // _automaton.dump();

        preparePatch();

        return _patch;
    }

    void initializeBad()
    {
        _bad = _automaton.getAcceptingStates();

        printf( "Initial bad states are:\n");
        for ( unsigned i = 0; i < _bad.size(); ++i )
        {
            printf( "%s\n", _bad[i].ascii() );
        }
        printf( "\n" );
    }

    void initializePredecessors()
    {
        for ( unsigned i = 0; i < _bad.size(); ++i )
        {
            Vector<String> predecessors = _automaton.allPredecessors( _bad[i] );

            for ( unsigned j = 0; j < predecessors.size(); ++j )
            {
                if ( !_bad.exists( predecessors[j] ) )
                {
                    _badFrontier.append( _bad[i] );
                    _predecessorsOfBad.append( predecessors[j] );
                }
            }
        }

        printf( "Initial predecessor states: \n" );
        for ( unsigned i = 0; i < _predecessorsOfBad.size(); ++i )
        {
            printf( "%s\n", _predecessorsOfBad[i].ascii() );
        }
        printf( "\n" );
    }

    void preparePatch()
    {
        for ( unsigned i = 0; i < _predecessorsOfBad.size(); ++i )
        {
            patchSingleState( _predecessorsOfBad[i] );
        }

        _patch.setNfa( _automaton );
    }

    void patchSingleState( String &predecessorState )
    {
        Vector<Event> eventsToBlock = eventsLeadingToBadStates( predecessorState );

        for ( unsigned i = 0; i < eventsToBlock.size(); ++i )
        {
            _patch.addBlockedEvent( predecessorState, eventsToBlock[i] );
        }
    }

    bool edgeLeadsToBadStates( const String &source, const NFA::Edge edge )
    {
        return ( edge._source == source ) && ( _bad.exists( edge._target ) );
    }

    static Event letterToEvent( const String &letter )
    {
        return Event( atoi( letter.ascii() ) );
    }

    static String eventToLetter( const Event &event )
    {
        return Stringf( "%u", event.type() );
    }

    bool needToRefineAPredecessor()
    {
        //        printf( "Need to refine a predecessor?\n" );

        for ( _predecessorIndex = 0; _predecessorIndex < _predecessorsOfBad.size(); ++_predecessorIndex )
        {
            // printf( "Checking: %s\n", _predecessorsOfBad[_predecessorIndex].ascii() );

            Vector<Event> blockingCandidates = eventsLeadingToBadStates( _predecessorsOfBad[_predecessorIndex] );

            // printf( "Events that need to be blocked are:\n" );
            // for ( unsigned i = 0; i < blockingCandidates.size(); ++i )
            // {
            //     printf( "%u ", blockingCandidates[i].type() );
            // }
            // printf( "\n" );

            printf( "\nChecking if refinement is required\n" );
            if ( goodEdgesGetBlocked( blockingCandidates ) || deadlocksAreCaused( blockingCandidates ) )
            {
                return true;
            }
            printf( "Refinement not required\n" );
        }

        return false;
    }

    EventVector eventsLeadingToBadStates( const String &state )
    {
        EventVector results;

        for ( unsigned i = 0; i < _automaton.edgeCount(); ++i )
        {
            if ( edgeLeadsToBadStates( state, _automaton.getEdge( i ) ) )
            {
                results.append( letterToEvent( _automaton.getEdge( i )._letter ) );
            }
        }

        results.uniqueElements();

        return results;
    }

    bool goodEdgesGetBlocked( Vector<Event> &blockingCandidates )
    {
        for ( unsigned i = 0; i < blockingCandidates.size(); ++i )
        {
            Vector<String> successors = _automaton.getAllSuccessors( _predecessorsOfBad[_predecessorIndex], eventToLetter( blockingCandidates[i] ) );

            for ( unsigned j = 0; j < successors.size(); ++j )
            {
                if ( !_bad.exists( successors[j] ) )
                {
                    printf( "Good edges will get blocked!\n" );
                    printf( "Problematic state: %s\n", _predecessorsOfBad[_predecessorIndex].ascii() );
                    printf( "Event: %u\n", blockingCandidates[i].type() );
                    printf( "Good state that will get cut off: %s\n", successors[j].ascii() );
                    return true;
                }
            }
        }

        //        printf( "good edges wont get blocked\n" );
        return false;
    }

    static String isolatePropertySuffix( const String &state )
    {
        char delimiter[] = "/";
        return state.tokenize( delimiter )[1];
    }

    bool deadlocksAreCaused( Vector<Event> &blockingCandidates )
    {
        String abstractComposedState = _predecessorsOfBad[_predecessorIndex];

        // TS_TRACE( Stringf( "Asking for concrete states of state: %s", abstractComposedState.ascii() ).ascii() );

        Vector<String> concreteComposedStates = _directAbstraction->abstractToConcrete( abstractComposedState );

        // TS_TRACE( Stringf( "Number of concrete states: %u", concreteComposedStates.size() ).ascii() );
        // TS_TRACE( "Asking for concrete states -- DONE" );

        // TS_TRACE( "Blocking candidates are:" );

        // for ( unsigned i = 0; i < blockingCandidates.size(); ++i )
        // {
        //     printf( "%u ", blockingCandidates[i].type() );
        // }
        // printf( "\n" );

        // printf( "Concrete composed states are:\n" );

        for ( unsigned i = 0; i < concreteComposedStates.size(); ++i )
        {
            ThreadState concreteState = _directAbstraction->constructStateFromLabel( concreteComposedStates[i] );
            // printf( "%s\t\t( ", concreteComposedStates[i].ascii() );

            // for ( unsigned j = 0; j < concreteState._blockedEvents.size(); ++j )
            // {
            //     printf( "%u ", concreteState._blockedEvents[j].type() );
            // }
            // printf( ")\n" );

            if ( concreteState._blockedEvents == _allEvents )
            {
                // printf( "\t A concrete deadlocked state, non related to patch. Assume its unreachable. Ignoring.\n" );
            }
            else
            {
                Vector<Event> eventsToBeBlocked = concreteState._blockedEvents + blockingCandidates;
                eventsToBeBlocked.uniqueElements();

                if ( eventsToBeBlocked == _allEvents )
                {
                    printf( "Deadlocks will be caused!\n" );
                    return true;
                }
            }
        }

        // printf( "Deadlocks wont be caused\n" );
        return false;
    }

    void refineAPredecessor()
    {
        _newBadState = "";
        _newPredecessors.clear();

        String predecessorToRefine = _predecessorsOfBad[_predecessorIndex];
        printf( "Refining predecessor: %s\n", predecessorToRefine.ascii() );

        EventVector eventsToBlock = eventsLeadingToBadStates( _predecessorsOfBad[_predecessorIndex] );

        Map<EventVector, StateSet> concreteStatesByEventSubset;

        classifyStatesToBlockSets( predecessorToRefine, eventsToBlock, concreteStatesByEventSubset );

        // TS_TRACE( "Done Classifying. Dumping classification" );

        // Map<EventVector, StateSet>::iterator it = concreteStatesByEventSubset.begin();

        // while ( it != concreteStatesByEventSubset.end() )
        // {
        //     printf( "Event Set: " );
        //     it->first.dump();

        //     printf( "\nStates:\n" );
        //     for ( unsigned i = 0; i < it->second.size(); ++i )
        //     {
        //         printf( "\t%s\n", it->second[i].ascii() );
        //     }

        //     ++it;
        // }

        //Map<StateSet, String> refinement;
        Vector<DirectAbstraction::AbstractStateData> refinement;

        createRefinement( concreteStatesByEventSubset, refinement, eventsToBlock, predecessorToRefine );

        // TS_TRACE( "Refinement created. Dumping." );

        // for ( unsigned i = 0; i < refinement.size(); ++i )
        // {
        //     refinement[i].dump();
        // }

        // TS_TRACE( "Removing original (refined) state" );

        _predecessorsOfBad.erase( predecessorToRefine );

        // TS_TRACE( "Updating direct abstraction" );

        _directAbstraction->refineState( predecessorToRefine, refinement );

        // TS_TRACE( "Adjusting the frontier if needed" );

        adjustFrontierAndPredecessors();
    }

    void adjustFrontierAndPredecessors()
    {
        if ( _newBadState != "" )
        {
            _bad.append( _newBadState );

            if ( stateShouldBeInTheFrontier( _newBadState ) )
            {
                _badFrontier.append( _newBadState );
                addPredecessorsToPre( _newBadState );
            }
        }

        for ( unsigned i = 0; i < _newPredecessors.size(); ++i )
        {
            _predecessorsOfBad.append( _newPredecessors[i] );
            addBadSuccessorsToFrontier( _newPredecessors[i] );
        }

        cleanFrontier();
    }

    void cleanFrontier()
    {
        UniqueVector<String>::iterator it = _badFrontier.begin();

        while ( it != _badFrontier.end() )
        {
            if ( !stateShouldBeInTheFrontier( *it ) )
            {
                // printf( "Removing a state from the frontier: %s\n", it->ascii() );
                _badFrontier.erase( it );
            }
            else
            {
                ++it;
            }
        }
    }

    bool stateShouldBeInTheFrontier( const String &state )
    {
        Vector<String> predecessors = _automaton.allPredecessors( state );

        for ( unsigned i = 0; i < predecessors.size(); ++i )
        {
            if ( !_bad.exists( predecessors[i] ) )
            {
                return true;
            }
        }

        return false;
    }

    void addPredecessorsToPre( const String &state )
    {
        Vector<String> predecessors = _automaton.allPredecessors( state );

        for ( unsigned i = 0; i < predecessors.size(); ++i )
        {
            if ( !_bad.exists( predecessors[i] ) )
            {
                _predecessorsOfBad.append( predecessors[i] );
            }
        }
    }

    void addBadSuccessorsToFrontier( const String &state )
    {
        Vector<String> successors = _automaton.allSuccessors( state );

        for ( unsigned i = 0; i < successors.size(); ++i )
        {
            if ( _bad.exists( successors[i] ) )
            {
                _badFrontier.append( successors[i] );
            }
        }
    }

    void createRefinement( Map<EventVector, StateSet> &concreteStatesByEventSubset,
                           Vector<DirectAbstraction::AbstractStateData> &refinement,
                           EventVector &eventsToBlock,
                           const String &baseName )
    {
        unsigned counter = 1;

        Map<EventVector, StateSet>::iterator it;

        for ( it = concreteStatesByEventSubset.begin(); it != concreteStatesByEventSubset.end(); ++it )
        {
            if ( it->first != eventsToBlock )
            {
                String newStateName = baseName + Stringf( "/%u", counter );
                ++counter;

                StateSet concreteStates = it->second;

                refinement.append( DirectAbstraction::AbstractStateData( concreteStates, newStateName ) );

                if ( !( it->first ).empty() )
                {
                    _newPredecessors.append( newStateName );
                }
            }
        }

        StateSet blockAllEvents;
        StateSet deadlockedStates;

        if ( concreteStatesByEventSubset.exists( eventsToBlock ) )
        {
            for ( unsigned i = 0; i < concreteStatesByEventSubset[eventsToBlock].size(); ++i )
            {
                ThreadState concreteState = _directAbstraction->constructStateFromLabel( concreteStatesByEventSubset[eventsToBlock][i] );
                Vector<Event> eventsToBeBlocked = concreteState._blockedEvents + eventsToBlock;
                eventsToBeBlocked.uniqueElements();

                if ( eventsToBeBlocked == _allEvents )
                {
                    deadlockedStates.append( concreteStatesByEventSubset[eventsToBlock][i] );
                }
                else
                {
                    blockAllEvents.append( concreteStatesByEventSubset[eventsToBlock][i] );
                }
            }
        }

        if ( !deadlockedStates.empty() )
        {
            String newStateName = baseName + String( "/bad" );
            refinement.append( DirectAbstraction::AbstractStateData( deadlockedStates, newStateName ) );
            _newBadState = newStateName;
        }
        if ( !blockAllEvents.empty() )
        {
            String newStateName = baseName + Stringf( "/%u", counter );
            refinement.append( DirectAbstraction::AbstractStateData( blockAllEvents, newStateName ) );
            _newPredecessors.append( newStateName );
        }
    }

    void classifyStatesToBlockSets( String predecessorToRefine, EventVector &eventsToBlock, Map<EventVector, StateSet> &concreteStatesByEventSubset )
    {
        // TS_TRACE( "Classifying concrete states to block sets" );

        Vector<String> allConcreteStates = _directAbstraction->abstractToConcrete( predecessorToRefine );

        for ( unsigned i = 0; i < allConcreteStates.size(); ++i )
        {
            // printf( "Concrete state: %s\n", allConcreteStates[i].ascii() );

            String concreteState = allConcreteStates[i];
            EventVector toBlockInState = determineEventsToBlockInState( concreteState, eventsToBlock );

            // printf( "Events to block in state are: " );
            // toBlockInState.dump();
            // printf( "\n" );

            if ( !concreteStatesByEventSubset.exists( toBlockInState ) )
            {
                StateSet newStateSet;
                concreteStatesByEventSubset[toBlockInState] = newStateSet;
            }

            concreteStatesByEventSubset[toBlockInState].append( concreteState );
        }
    }

    EventVector determineEventsToBlockInState( const String &concreteState, const EventVector &eventsToBlock )
    {
        EventVector results;

        for ( unsigned i = 0; i < eventsToBlock.size(); ++i )
        {
            if ( eventLeadsToBadState( concreteState, eventsToBlock.get( i ) ) )
            {
                results.append( eventsToBlock.get( i ) );
            }
        }

        return results;
    }

    bool eventLeadsToBadState( const String &concreteState, const Event &event )
    {
        ThreadState threadState = _directAbstraction->constructStateFromLabel( concreteState );

        // printf( "\tState: %s\n", concreteState.ascii() );

        // printf( "\tRequested events: " );

        // for ( unsigned i = 0; i < threadState._requestedEvents.size(); ++i )
        // {
        //     printf( " %u", threadState._requestedEvents[i].type() );
        // }

        // printf( "\n" );

        // printf( "\tBlocked events: " );

        // for ( unsigned i = 0; i < threadState._blockedEvents.size(); ++i )
        // {
        //     printf( " %u", threadState._blockedEvents[i].type() );
        // }

        // printf( "\n" );

        if ( threadState._blockedEvents.exists( event ) )
        {
            // printf( "Event %u is already blocked. Return false.\n", event.type() );
            return false;
        }

        String concreteSuccessor = _directAbstraction->findConcreteSuccessor( concreteState, event );

        // printf( "Cocnrete successor on event %u is: %s\n", event.type(), concreteSuccessor.ascii() );
        String abstractSuccessor = _directAbstraction->concreteToAbstract( concreteSuccessor );

        // printf( "And its abstraction is: %s\n", abstractSuccessor.ascii() );
        return _bad.exists( abstractSuccessor );
    }

    void dumpFrontier()
    {
        static unsigned iteration = 0;

        TS_TRACE( Stringf( "Dumping predecessors of bad: iteration %u", iteration ).ascii() );

        for ( unsigned i = 0; i < _predecessorsOfBad.size(); ++i )
        {
            printf( "State: %s\n", _predecessorsOfBad[i].ascii() );

            Vector<String> concrete = _directAbstraction->abstractToConcrete( _predecessorsOfBad[i] );

            for ( unsigned j = 0; j < concrete.size(); ++j )
            {
                printf( "\t%s\n", concrete[j].ascii() );
            }

        }

        ++iteration;
    }
};

#endif // __AbstractSafetyPatcher_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
