#ifndef __CegarRefiner_h__
#define __CegarRefiner_h__

#include "AbstractStateSplitter.h"
#include "AbstractionData.h"
#include "NFA.h"
#include "Vector.h"

class CegarRefiner
{
public:
    bool checkAndRefine( NFA &productAutomaton, NFA::Run &counterExample, AbstractionData &abstractionData, DirectAbstraction &directAbstraction )
    {
        initialize( productAutomaton, counterExample, abstractionData, directAbstraction );

        if ( checkIfSpurious() )
        {
            TS_TRACE( "Exmaple is spurious --- refining" );
            refine();
            return true;
        }
        else
        {
            return false;
        }
    }

private:
    NFA *_productAutomaton;
    NFA::Run *_counterExample;
    AbstractionData *_abstractionData;
    DirectAbstraction *_directAbstraction;
    unsigned _numberOfModules;
    unsigned _indexInRun;

    typedef UniqueVector<String> StateSet;

    Vector<StateSet> _reachableConcreteStates;

    void initialize( NFA &productAutomaton, NFA::Run &counterExample, AbstractionData &abstractionData, DirectAbstraction &directAbstraction )
    {
        if ( counterExample.empty() )
        {
            throw Error( Error::REFINING_FOR_AN_EMPTY_RUN );
        }

        _productAutomaton = &productAutomaton;
        _counterExample = &counterExample;
        _abstractionData = &abstractionData;
        _directAbstraction = &directAbstraction;

        _numberOfModules = _abstractionData->_abstractModules.size();
        _indexInRun = 0;
    }

    bool checkIfSpurious()
    {
        TS_TRACE( "Beginning to check if the example is spurious" );

        createInitialSet();
        dumpReachableStates();

        if ( previousSetIsEmpty() )
        {
            return true;
        }

        for ( _indexInRun = 1; _indexInRun < _counterExample->_states.size(); ++_indexInRun )
        {
            calculateNewSet();

            if ( previousSetIsEmpty() )
            {
                return true;
            }

            dumpReachableStates();
        }

        return false;
    }

    void refine()
    {
        String stateToRefine = _counterExample->_states[_indexInRun-1];

        // printf( "Index in run: %u\n", _indexInRun );
        printf( "State to refine is: %s\n", stateToRefine.ascii() );

        Vector<String> concreteStates = _directAbstraction->abstractToConcrete( stateToRefine );

        // printf( "All concrete states mapped to the state are:\n" );
        // for ( unsigned i = 0; i < concreteStates.size(); ++i )
        // {
        //     printf( "\t%s\n", concreteStates[i].ascii() );
        // }

        Vector<String> concreteStatesToSeparate;

        if ( _indexInRun > 1 )
        {
            concreteStatesToSeparate = _reachableConcreteStates[_reachableConcreteStates.size()-2];
            // TS_TRACE( concreteStatesToSeparate.size() );
        }
        else
        {
            concreteStatesToSeparate.append( findConcreteInitialState() );
        }

        printf( "Concrete states to separate are:\n" );
        for ( unsigned i = 0; i < concreteStatesToSeparate.size(); ++i )
        {
            concreteStatesToSeparate[i] += String( "/" ) + isolatePropertySuffix( stateToRefine );
            concreteStates.eraseByValue( concreteStatesToSeparate[i] );
            printf( "\t%s\n", concreteStatesToSeparate[i].ascii() );
        }

        DirectAbstraction::AbstractStateData newAbstractState( concreteStatesToSeparate, stateToRefine + "/1" );
        DirectAbstraction::AbstractStateData complementState( concreteStates, stateToRefine + "/2" );

        Vector<DirectAbstraction::AbstractStateData> refinement;
        refinement.append( newAbstractState );
        refinement.append( complementState );

        _directAbstraction->refineState( stateToRefine, refinement );
    }

    String findConcreteInitialState()
    {
        String concreteInitialState = _abstractionData->_abstractModules[0]._concrete->_initialState;
        for ( unsigned i = 1; i < _numberOfModules; ++i )
        {
            concreteInitialState += String( " " ) + _abstractionData->_abstractModules[i]._concrete->_initialState;
        }

        return concreteInitialState;
    }

    void createInitialSet()
    {
        String concreteInitialState = findConcreteInitialState();

        if ( stateBelongsToCounterExample( concreteInitialState ) )
        {
            StateSet initiaStates;
            initiaStates.append( concreteInitialState );
            _reachableConcreteStates.append( initiaStates );
        }
    }

    static Event letterToEvent( const String &letter )
    {
        return Event( atoi( letter.ascii() ) );
    }

    bool previousSetIsEmpty() const
    {
        return _reachableConcreteStates.last().empty();
    }

    void calculateNewSet()
    {
        // TS_TRACE( "Calculate new set called" );

        Event event = letterToEvent( _counterExample->_letters[_indexInRun - 1] );
        StateSet newConcreteStates;

        for ( unsigned i = 0; i < _reachableConcreteStates.last().size(); ++i )
        {
            // TS_TRACE( Stringf( "Previous state: %s\n", _reachableConcreteStates.last()[i].ascii() ).ascii() );
            findConcreteSuccessor( _reachableConcreteStates.last()[i], event, newConcreteStates );
        }

        // TS_TRACE( Stringf( "Number of concrete successors found: %u", newConcreteStates.size() ).ascii() );

        StateSet::iterator it = newConcreteStates.begin();

        while ( it != newConcreteStates.end() )
        {
            if ( !stateBelongsToCounterExample( *it ) )
            {
                newConcreteStates.erase( it );
            }
            else
            {
                ++it;
            }
        }

        _reachableConcreteStates.append( newConcreteStates );
    }

    bool stateBelongsToCounterExample( String concreteState )
    {
        concreteState += String( "/" ) +  isolatePropertySuffix( _counterExample->_states[_indexInRun] );
        // printf( "Checking if state belongs to counter example. State: %s\n", concreteState.ascii() );

        return _directAbstraction->concreteToAbstract( concreteState ) == _counterExample->_states[_indexInRun];
    }

    void findConcreteSuccessor( String stateLabel, Event event, StateSet &successorStates )
    {
        ThreadState state = _directAbstraction->constructStateFromLabel( stateLabel );

        if ( !state._requestedEvents.exists( event ) )
        {
            return;
        }

        TS_ASSERT_EQUALS( state.labelsOfSuccessors( event ).size(), 1U );
        String successorState = state.labelsOfSuccessors( event ).first();
        successorStates.append( successorState );
    }

    void dumpReachableStates()
    {
        printf( "Dumping reachable concrete states --- iteration %u\n", _indexInRun );

        if ( _reachableConcreteStates[_indexInRun].empty() )
        {
            printf( "\tSet is empty!\n\n" );
            return;
        }

        for ( unsigned i = 0; i < _reachableConcreteStates[_indexInRun].size(); ++i )
        {
            printf( "\t%s\n", _reachableConcreteStates[_indexInRun][i].ascii() );
        }

        printf( "\n" );
    }

    static String isolatePropertySuffix( const String &state )
    {
        char delimiter[] = "/";
        return state.tokenize( delimiter )[1];
    }

    static String removePropertySuffix( const String &state )
    {
        char delimiter[] = "/";
        return state.tokenize( delimiter )[0];
    }
};

#endif // __CegarRefiner_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
