#ifndef __NFA_h__
#define __NFA_h__

#include "Map.h"
#include "Queue.h"
#include "String.h"
#include "Vector.h"

class NFA
{
public:
    struct Run
    {
        Vector<String> _letters;
        Vector<String> _states;

        bool empty() const
        {
            return _states.empty();
        }

        void clear()
        {
            _letters.clear();
            _states.clear();
        }
    };

    struct Edge
    {
        Edge( const String &source, const String &letter, const String &target ) : _source( source ), _letter( letter ), _target( target )
        {
        }

        String _source;
        String _letter;
        String _target;

        bool operator==( const Edge &other )
        {
            return ( _source == other._source ) && ( _letter == other._letter ) && ( _target == other._target );
        }
    };

    void addLetter( const String &letter )
    {
        _alphabet.append( letter );
    }

    void addState( const String &state )
    {
        _states.append( state );
    }

    void removeState( const String &state )
    {
        _states.erase( state );
    }

    void addEdge( Edge edge )
    {
        _edges.append( edge );
    }

    unsigned edgeCount()
    {
        return _edges.size();
    }

    Edge getEdge( unsigned index )
    {
        return _edges[index];
    }

    void setInitialState( const String &initialState )
    {
        _initialState = initialState;
    }

    String getInitialState() const
    {
        return _initialState;
    }

    void markAsAccepting( const String &acceptingState )
    {
        _acceptingStates.append( acceptingState );
    }

    unsigned numberOfStates() const
    {
        return _states.size();
    }

    String getState( unsigned i )
    {
        return _states[i];
    }

    bool acceptingState( const String &state )
    {
        return _acceptingStates.exists( state );
    }

    bool existsSuccessor( const String &state, const String &letter )
    {
        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            if ( ( _edges[i]._source == state ) && ( _edges[i]._letter == letter ) )
            {
                return true;
            }
        }

        return true;
    }

    String getSuccessor( const String &state, const String &letter )
    {
        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            if ( ( _edges[i]._source == state ) && ( _edges[i]._letter == letter ) )
            {
                return _edges[i]._target;
            }
        }

        //printf( "State: %s. Letter: %s\n", state.ascii(), letter.ascii() );
        throw Error( Error::NO_SUCCESSOR_IN_NFA );
    }

    Vector<String> getAllSuccessors( const String &state, const String &letter )
    {
        UniqueVector<String> successors;

        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            if ( ( _edges[i]._source == state ) && ( _edges[i]._letter == letter ) )
            {
                successors.append( _edges[i]._target );
            }
        }

        return successors;
    }

    void dump()
    {
        printf( "Dumping NFA:\n" );

        printf( "Number of letters: %u\nLetters are:", _alphabet.size() );
        for ( unsigned i = 0; i < _alphabet.size() - 1; ++i )
        {
            printf( "%s, ", _alphabet[i].ascii() );
        }
        printf( "%s\n\n", _alphabet[_alphabet.size() - 1].ascii() );

        printf( "Number of states: %u\n", _states.size() );
        for ( unsigned i = 0; i < _states.size(); ++i )
        {
            printf( "%s\n", _states[i].ascii() );
        }

        printf( "Initial state: %s\n\n", _initialState.ascii() );

        printf( "Number of accepting states: %u\n", _acceptingStates.size() );
        for ( unsigned i = 0; i < _acceptingStates.size(); ++i )
        {
            printf( "%s\n", _acceptingStates[i].ascii() );
        }

        printf( "\nNumber of edges: %u\n", _edges.size() );
        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            printf( "(%s, %s, %s)\n", _edges[i]._source.ascii(), _edges[i]._letter.ascii(), _edges[i]._target.ascii() );
        }
    }

    bool emptyLanguage() const
    {
        return _acceptingStates.empty();
    }

    NFA::Run findRunInLanguage()
    {
        runBfs();

        String state = reachableAcceptingState();

        NFA::Run run;

        while ( predecessor( state ) != "" )
        {
            if ( acceptingState( predecessor( state ) ) )
            {
                run.clear();
            }
            else
            {
                run._letters.insertHead( letterLeadingToState( state ) );
                run._states.insertHead( state );
            }

            state = predecessor( state );
        }

        if ( acceptingState( predecessor( state ) ) )
        {
            run.clear();
        }

        run._states.insertHead( state );

        return run;
    }

    Vector<String> getAcceptingStates() const
    {
        return _acceptingStates;
    }

    void eliminateNonReachables()
    {
        runBfs();
        removeNonReachableStates();
    }

    UniqueVector<String> allPredecessors( const String &state )
    {
        UniqueVector<String> results;

        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            if ( _edges[i]._target == state )
            {
                results.append( _edges[i]._source );
            }
        }

        return results;
    }

    UniqueVector<String> allSuccessors( const String &state )
    {
        UniqueVector<String> results;

        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            if ( _edges[i]._source == state )
            {
                results.append( _edges[i]._target );
            }
        }

        return results;
    }

    void clear()
    {
        _states.clear();
        _alphabet.clear();
        _edges.clear();
        _initialState = "ERROR";
        _acceptingStates.clear();

        _statesToVisit.clear();
        _visitedStates.clear();
    }

    void unifyAcceptingStates( const String &newStateName )
    {
        Vector<Edge>::iterator it = _edges.begin();

        while ( it != _edges.end() )
        {
            if ( _acceptingStates.exists( it->_target ) )
            {
                it->_target = newStateName;
            }

            if ( _acceptingStates.exists( it->_source ) )
            {
                _edges.erase( it );
            }
            else
            {
                ++it;
            }
        }

        if ( _acceptingStates.exists( _initialState ) )
        {
            _initialState = newStateName;
        }

        for ( unsigned i = 0; i < _acceptingStates.size(); ++i )
        {
            _states.erase( _acceptingStates[i] );
        }

        _states.append( newStateName );

        _acceptingStates.clear();
        _acceptingStates.append( newStateName );
    }

    Vector<Edge> &edges()
    {
        return _edges;
    }

private:
    Vector<String> _states;
    Vector<String> _alphabet;
    Vector<Edge> _edges;
    String _initialState;
    Vector<String> _acceptingStates;

    Queue<String> _statesToVisit;
    Map<String, String> _visitedStates;

    void runBfs()
    {
        initBfs();

        while ( haveStatesToVisit() )
        {
            visitSingleState();
        }
    }

    void initBfs()
    {
        _statesToVisit.clear();
        _visitedStates.clear();

        _statesToVisit.push( _initialState );
        _visitedStates[_initialState] = "";
    }

    bool haveStatesToVisit() const
    {
        return !_statesToVisit.empty();
    }

    void visitSingleState()
    {
        String state = _statesToVisit.peak();
        _statesToVisit.pop();

        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            if ( ( _edges[i]._source == state ) && ( !_visitedStates.exists( _edges[i]._target ) ) )
            {
                _visitedStates[_edges[i]._target] = state;
                _statesToVisit.push( _edges[i]._target );
            }
        }
    }

    void removeNonReachableStates()
    {
        Vector<String>::iterator it = _states.begin();

        while ( it != _states.end() )
        {
            if ( !_visitedStates.exists( *it ) )
            {
                removeEdgesOfState( *it );
                removeFromAcceptingStates( *it );
                _states.erase( it );
            }
            else
            {
                ++it;
            }
        }
    }

    void removeFromAcceptingStates( String &state )
    {
        if ( acceptingState( state ) )
        {
            _acceptingStates.erase( state );
        }
    }

    void removeEdgesOfState( String &state )
    {
        Vector<Edge>::iterator it = _edges.begin();

        while ( it != _edges.end() )
        {
            if ( ( it->_source == state ) || ( it->_target == state ) )
            {
                _edges.erase( it );
            }
            else
            {
                ++it;
            }
        }
    }

    String reachableAcceptingState()
    {
        if ( _acceptingStates.empty() )
        {
            throw Error( Error::NO_REACHABLE_ACCEPTING_STATES );
        }

        return _acceptingStates[0];
    }

    String predecessor( String &state )
    {
        return _visitedStates[state];
    }

    String letterLeadingToState( String &state )
    {
        String predecessorState = predecessor( state );

        for ( unsigned i = 0; i < _edges.size(); ++i )
        {
            if ( ( _edges[i]._source == predecessorState ) && ( _edges[i]._target == state ) )
            {
                return _edges[i]._letter;
            }
        }

        throw Error( Error::NO_REACHABLE_ACCEPTING_STATES );
    }
};

#endif // __NFA_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
