#ifndef __ThreadStateGraph_h__
#define __ThreadStateGraph_h__

#include "Map.h"
#include "Queue.h"
#include "Stringf.h"
#include "ThreadState.h"

struct ThreadStateGraph
{
    Map<String, ThreadState> _graph;
    Vector<Event> _events;
    String _initialState;
    String _threadName;
    AtomicPropositions _atomicPropositions;

    void clear()
    {
        _graph.clear();
        _events.clear();
        _initialState = "";
        _threadName = "";
        _atomicPropositions.clear();
    }

    void applyAxioms()
    {
        removeNonRequested();
        eliminateNonReachables();
    }

    // void assertNoDeadlocks()
    // {
    //     Map<String, ThreadState>::iterator it;

    //     for ( it = _graph.begin(); it != _graph.end(); ++it )
    //     {
    //         TS_ASSERT( !it->second._successors.empty() );
    //     }
    // }

    void dump()
    {
        printf( "State graph size is: %u\n\n", _graph.size() );

        printf( "Initial state: %s\n", _initialState.ascii() );

        printf( "Atomic propositions of the graph:\n" );
        for ( unsigned i = 0; i < _atomicPropositions.size(); ++i )
        {
            printf( "%s ", _atomicPropositions.at( i ).ascii() );
        }
        printf( "\n\n" );

        Map<String, ThreadState>::iterator it;
        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            printf( "%s:\n", it->second._label.ascii() );

            printf( "Requested: " );
            for ( unsigned i = 0; i < it->second._requestedEvents.size(); ++i )
            {
                printf( "%u ", it->second._requestedEvents[i].type() );
            }
            printf( "\n" );

            printf( "Blocked: " );
            for ( unsigned i = 0; i < it->second._blockedEvents.size(); ++i )
            {
                printf( "%u ", it->second._blockedEvents[i].type() );
            }
            printf( "\n" );

            // printf( "History: " );
            // for ( unsigned i = 0; i < it->second._history.size(); ++i )
            // {
            //     printf( "%u ", it->second._history[i].type() );
            // }
            // printf( "\n" );

            printf( "Successors: " );
            Vector<unsigned> selfLoops;

            for ( unsigned i = 0; i < it->second._successors.size(); ++i )
            {
                if ( it->second._successors[i]._state != it->second._label )
                {
                    printf( "%u-->%s  ", it->second._successors[i]._event.type(), it->second._successors[i]._state.ascii() );
                }
                else
                {
                    selfLoops.append( it->second._successors[i]._event.type() );
                }
            }
            printf( "\n" );

            printf( "Self Loops: " );
            for ( unsigned i = 0; i < selfLoops.size(); ++i )
            {
                printf( "%u ", selfLoops[i] );
            }
            printf( "\n" );

            // printf( "Predecessors: " );
            // for ( unsigned i = 0; i < it->second._predecessors.size(); ++i )
            // {
            //     printf( "%u-->%s  ", it->second._predecessors[i]._event.type(),
            //             it->second._predecessors[i]._state.ascii() );
            // }
            // printf( "\n" );

            printf( "Atomic Propositions: " );
            for ( unsigned i = 0; i < it->second._atomicPropositions.size(); ++i )
            {
                printf( "%s ", it->second._atomicPropositions.at( i ).ascii() );
            }
            printf( "\n\n" );
        }
    }

    void eliminateNonReachables()
    {
        init();

        while ( moreStatesToCheck() )
        {
            checkAnotherState();
        }

        Vector<String> toErase;

        Map<String, ThreadState>::iterator it;

        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            if ( !_visitedStates.exists( it->first ) )
            {
                toErase.append( it->first );
            }
        }

        for ( unsigned i = 0; i < toErase.size(); ++i )
        {
            removeState( toErase[i] );
        }
    }

    void removeState( const String &label )
    {
        // TS_TRACE( Stringf( "Removing state: %s", label.ascii() ).ascii() );

        for ( unsigned i = 0; i < _graph[label]._successors.size(); ++i )
        {
            String successor = _graph[label]._successors[i]._state;
            // TS_TRACE( Stringf( "Handling successor: %s", successor.ascii() ).ascii() );
            _graph[successor].removePredecessor( label );
        }

        for ( unsigned i = 0; i < _graph[label]._predecessors.size(); ++i )
        {
            String predecessor = _graph[label]._predecessors[i]._state;
            // TS_TRACE( Stringf( "Handling predecessor: %s", predecessor.ascii() ).ascii() );
            _graph[predecessor].removeSuccessor( label );
        }

        _graph.erase( label );
    }

private:
    Queue<String> _statesToCheck;
    Map<String, char> _visitedStates;

    void init()
    {
        _visitedStates.clear();
        _statesToCheck.clear();

        _visitedStates[_initialState] = 0;
        _statesToCheck.push( _initialState );
    }

    bool moreStatesToCheck() const
    {
        return !_statesToCheck.empty();
    }

    void checkAnotherState()
    {
        Vector<String> successors = _graph[_statesToCheck.peak()].allSuccessors();
        _statesToCheck.pop();

        for ( unsigned i = 0; i < successors.size(); ++i )
        {
            if ( !_visitedStates.exists( successors[i] ) )
            {
                _visitedStates[successors[i]] = 0;
                _statesToCheck.push( successors[i] );
            }
        }
    }

    void addAllUnreachedStates()
    {
        Map<String, ThreadState>::iterator it;

        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            if ( !_visitedStates.exists( it->first ) )
            {
                _visitedStates[it->first] = 0;
            }
        }
    }

    void removeNonRequested()
    {
        Map<String, ThreadState>::iterator it;

        for ( it = _graph.begin(); it != _graph.end(); ++it )
        {
            removeNonRequstedAtState( it->first );
        }
    }

    void removeNonRequstedAtState( const String &label )
    {
        unsigned i = 0;

        while ( i < _graph[label]._successors.size() )
        {
            if ( !_graph[label]._requestedEvents.exists( _graph[label]._successors[i]._event ) )
            {
                removeTransition( label, i );
            }
            else
            {
                ++i;
            }
        }
    }

    void removeTransition( const String &label, unsigned successorIndex )
    {
        String successor = _graph[label]._successors[successorIndex]._state;

        Transition predecessor( _graph[label]._successors[successorIndex]._event, label );
        _graph[successor]._predecessors.erase( predecessor );

        _graph[label]._successors.eraseAt( successorIndex );
    }
};

#endif // __ThreadStateGraph_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
