#include <cxxtest/TestSuite.h>

#include "AbstractSafetyPatcher.h"
#include "CegarRefiner.h"
#include "Event.h"
#include "FairnessApproximator.h"
#include "FairnessPatchGenerator.h"
#include "ForkOneSpanner.h"
#include "ForkThreeSpanner.h"
#include "ForkTwoSpanner.h"
#include "GraphSimplifier.h"
#include "HotStateAnalyzer.h"
#include "IEvent.h"
#include "LivenessChecker.h"
#include "MinimalThreadAbstraction.h"
#include "MockErrno.h"
#include "MultipleThreadIntersector.h"
#include "NFA.h"
#include "PhilOneSpanner.h"
#include "PhilThreeSpanner.h"
#include "PhilTwoSpanner.h"
#include "RealMalloc.h"
#include "SafetyModelChecker.h"
#include "SafetyPatcher.h"
#include "ThreadPartitioner.h"
#include "Vector.h"

#include "safety/OneAndTwoDisjoint.h"
#include "safety/OneThriceInARow.h"
#include "safety/OneTwiceInARow.h"

class MockForSpan :
    public RealMalloc,
    public MockErrno,
    public T::Real_open,
    public T::Real_close,
    public T::Real_write,
    public T::Real_read
{
public:
};

class SpanTestSuite : public CxxTest::TestSuite
{
public:
    MockForSpan mock;

    ThreadStateGraph _philOne;
    ThreadStateGraph _philTwo;
    ThreadStateGraph _philThree;

    ThreadStateGraph _forkOne;
    ThreadStateGraph _forkTwo;
    ThreadStateGraph _forkThree;

    ThreadStateGraph _concreteProgram;
    MultipleThreadIntersector _intersector;
    AbstractionData _abstractionData;
    Vector<BModule> _bModules;

    SpanTestSuite()
    {
        _philOne._events = IEvent::allEvents();
        _philTwo._events = IEvent::allEvents();
        _philThree._events = IEvent::allEvents();

        _forkOne._events = IEvent::allEvents();
        _forkTwo._events = IEvent::allEvents();
        _forkThree._events = IEvent::allEvents();
    }

    void test_program_properties__concrete()
    {
        TS_TRACE( "Concrete program verification in progress..." );

        spanThreadGraphs();
        createFullProgramGraph();

        NFA property =
            // OneAndTwoDisjoint::getNfa()
            OneTwiceInARow::getNfa()
            // OneThriceInARow::getNfa()
            ;

        // concreteSafetyTest( property );
        concreteSafetyPatching( property );
    }

    // void test_program_properties__abstract()
    // {
    //     TS_TRACE( "Abstract program verification in progress..." );

    //     spanThreadGraphs();
    //     createAbstractGraph();

    //     NFA property =
    //         // OneAndTwoDisjoint::getNfa()
    //         OneTwiceInARow::getNfa()
    //         // OneThriceInARow::getNfa()
    //         ;

    //     // safetyTest( property );
    //     // cegarSafetyTest( property );
    //     safetyPatch( property );
    // }

    void spanThreadGraphs()
    {
        TS_TRACE( "\tSpanning thread graphs" );

        PhilOneSpanner().span( _philOne );
        PhilTwoSpanner().span( _philTwo );
        PhilThreeSpanner().span( _philThree );

        ForkOneSpanner().span( _forkOne );
        ForkTwoSpanner().span( _forkTwo );
        ForkThreeSpanner().span( _forkThree );
    }

    void createFullProgramGraph()
    {
        TS_TRACE( "\tIntersecting threads" );

        _intersector.addThread( &_philOne );
        _intersector.addThread( &_philTwo );
        _intersector.addThread( &_philThree );
        _intersector.addThread( &_forkOne );
        _intersector.addThread( &_forkTwo );
        _intersector.addThread( &_forkThree );

        _intersector.intersect( _concreteProgram, ' ' );

        TS_TRACE( Stringf( "Numebr of states before axioms: %u", _concreteProgram._graph.size() ).ascii() );

        _concreteProgram.applyAxioms();

        TS_TRACE( Stringf( "Numebr of states after axioms: %u", _concreteProgram._graph.size() ).ascii() );

        TS_TRACE( "Full program:" );
        _concreteProgram.dump();
    }

    void createAbstractGraph()
    {
        partitionToModules();
        abstractEachModule();
        composeAbstractModules();
    }

    void partitionToModules()
    {
        BModule module1;
        module1.add( &_philOne );
        module1.add( &_forkOne );

        BModule module2;
        module2.add( &_philTwo );
        module2.add( &_forkTwo );
        module2.add( &_philThree );
        module2.add( &_forkThree );

        _bModules.append( module1 );
        _bModules.append( module2 );
    }

    void abstractEachModule()
    {
        for ( unsigned i = 0; i < _bModules.size(); ++i )
        {
            TS_TRACE( Stringf( "Abstracting module #%u", i ).ascii() );
            abstractSingleModule( _bModules[i] );
        }
    }

    void abstractSingleModule( BModule &module )
    {
        char moduleInnerSeparationChar = '.';

        MultipleThreadIntersector intersector;

        for ( unsigned i = 0; i < module._threads.size(); ++i )
        {
            intersector.addThread( module._threads[i] );
        }

        ThreadStateGraph *composedModule = new ThreadStateGraph;
        intersector.intersect( *composedModule, moduleInnerSeparationChar );
        TS_TRACE( Stringf( "Size of module before abstraction: %u", composedModule->_graph.size() ).ascii() );

        ThreadStateGraph *abstractModule = new ThreadStateGraph;
        AbstractionFunction *abstractionFunction = new AbstractionFunction;

        MinimalThreadAbstraction::abstract( *composedModule, *abstractModule, *abstractionFunction );
        TS_TRACE( Stringf( "Size of module after abstraction: %u", abstractModule->_graph.size() ).ascii() );

        _abstractionData._abstractModules.append( AbstractionData::ModuleData( composedModule, abstractModule, abstractionFunction ) );
    }

    void composeAbstractModules()
    {
        _abstractionData._composedGraph = new ThreadStateGraph;

        char separationCharBetweenModules = ' ';
        MultipleThreadIntersector intersector;

        for ( unsigned i = 0; i < _abstractionData._abstractModules.size(); ++i )
        {
            intersector.addThread( _abstractionData._abstractModules[i]._abstract );
        }

        intersector.intersect( *_abstractionData._composedGraph, separationCharBetweenModules );

        TS_TRACE( "Size of abstract program before finalization:" );
        TS_TRACE( _abstractionData._composedGraph->_graph.size() );

        TS_TRACE( "And after finalization:" );
        _abstractionData._composedGraph->applyAxioms();
        TS_TRACE( _abstractionData._composedGraph->_graph.size() );
    }

    void safetyTest( NFA &safetyProperty )
    {
        NFA::Run counterExample;
        NFA productAutomaton;

        try
        {
            if ( SafetyModelChecker::checkSafetyProperty( *_abstractionData._composedGraph, safetyProperty, counterExample, productAutomaton ) )
            {
                TS_TRACE( "Property holds!" );
            }
            else
            {
                TS_TRACE( "Property doesnt hold!" );
            }
        }
        catch ( const Error &e )
        {
            TS_TRACE( Stringf( "Safety test failed! Error: %u", e.code() ).ascii() );
        }
    }

    void concreteSafetyTest( NFA &safetyProperty )
    {
        NFA::Run counterExample;
        NFA productAutomaton;

        try
        {
            if ( SafetyModelChecker::checkSafetyProperty( _concreteProgram, safetyProperty, counterExample, productAutomaton ) )
            {
                TS_TRACE( "Property holds!" );
            }
            else
            {
                TS_TRACE( "Property doesnt hold!" );
            }
        }
        catch ( const Error &e )
        {
            TS_TRACE( Stringf( "Safety test failed! Error: %u", e.code() ).ascii() );
        }
    }

    void cegarSafetyTest( NFA &safetyProperty )
    {
        NFA productAutomaton;
        NFA::Run counterExample;
        DirectAbstraction directAbstraction( productAutomaton, _abstractionData );

        bool flag = false;

        while ( !flag )
        {
            counterExample.clear();

            if ( SafetyModelChecker::checkSafetyProperty( *_abstractionData._composedGraph, safetyProperty, counterExample, productAutomaton ) )
            {
                TS_TRACE( "Property Holds!" );
                flag = true;
            }
            else
            {
                TS_TRACE( "Property doesn't hold on the abstract graph" );

                dumpCounterExample( counterExample );

                CegarRefiner cegarRefiner;

                if ( !cegarRefiner.checkAndRefine( productAutomaton, counterExample, _abstractionData, directAbstraction ) )
                {
                    printf( "The counter example is genuine!\n" );
                    flag = true;
                }
            }
        }
    }

    void dumpCounterExample( NFA::Run &counterExample )
    {
        printf( "Counter example:\n" );
        for ( unsigned i = 0; i < counterExample._states.size() - 1; ++i )
        {
            printf( "%s\n\t(%s)\n", counterExample._states[i].ascii(), counterExample._letters[i].ascii() );
        }
        printf( "%s\n", counterExample._states.last().ascii() );
    }

    void safetyPatch( NFA &safetyProperty )
    {
        NFA::Run counterExample;
        NFA productAutomaton;

        try
        {
            if ( SafetyModelChecker::checkSafetyProperty( *_abstractionData._composedGraph, safetyProperty, counterExample, productAutomaton ) )
            {
                TS_TRACE( "Property holds!" );
                return;
            }
            else
            {
                TS_TRACE( "Property doesnt hold!" );
            }

            SafetyPatch patch;

            patch = AbstractSafetyPatcher::patch( productAutomaton, _abstractionData );
            patch.dump();
        }
        catch ( const Error &e )
        {
            TS_TRACE( Stringf( "Patching failed! Error: %u", e.code() ).ascii() );
        }
    }

    void concreteSafetyPatching( NFA &safetyProperty )
    {
        NFA::Run counterExample;
        NFA productAutomaton;

        try
        {
            if ( SafetyModelChecker::checkSafetyProperty( _concreteProgram, safetyProperty, counterExample, productAutomaton ) )
            {
                TS_TRACE( "Property holds!" );
            }
            else
            {
                TS_TRACE( "Property doesnt hold!" );
            }

            SafetyPatch patch = SafetyPatcher::patch( productAutomaton );
            patch.dump();
        }
        catch ( const Error &e )
        {
            TS_TRACE( Stringf( "Safety test failed! Error: %u", e.code() ).ascii() );
        }
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
